@@ -27,11 +27,13 @@ import (
 	"log"
 	"net/http"
 	"net/url"
+	"reflect"
 	"strconv"
 	"strings"
 	"sync"
 	"sync/atomic"
 	"testing"
+	"testing/iotest"
 	"time"
 
 	"github.com/google/go-cmp/cmp"
@@ -70,16 +72,23 @@ type HarnessMaker func(ctx context.Context, t *testing.T) (Harness, error)
 // 2. Creates a blob in a directory, using BeforeWrite as a WriterOption.
 // 3. Fetches the blob's attributes and calls AttributeCheck.
 // 4. Creates a Reader for the blob using BeforeReader as a ReaderOption,
-//    and calls ReaderCheck with the resulting Reader.
+//
+//	and calls ReaderCheck with the resulting Reader.
+//
 // 5. Calls List using BeforeList as a ListOption, with Delimiter set so
-//    that only the directory is returned, and calls ListObjectCheck
-//    on the single directory list entry returned.
+//
+//	that only the directory is returned, and calls ListObjectCheck
+//	on the single directory list entry returned.
+//
 // 6. Calls List using BeforeList as a ListOption, and calls ListObjectCheck
-//    on the single blob entry returned.
+//
+//	on the single blob entry returned.
+//
 // 7. Tries to read a non-existent blob, and calls ErrorCheck with the error.
 // 8. Makes a copy of the blob, using BeforeCopy as a CopyOption.
 // 9. Calls SignedURL using BeforeSign as a SignedURLOption for each supported
-//    signing method (i.e. GET, PUT and DELETE).
+//
+//	signing method (i.e. GET, PUT and DELETE).
 //
 // For example, an AsTest might set a driver-specific field to a custom
 // value in BeforeWrite, and then verify the custom value was returned in
@@ -210,6 +219,9 @@ func RunConformanceTests(t *testing.T, newHarness HarnessMaker, asTests []AsTest
 	t.Run("TestListDelimiters", func(t *testing.T) {
 		testListDelimiters(t, newHarness)
 	})
+	t.Run("TestDirsWithCharactersBeforeDelimiter", func(t *testing.T) {
+		testDirsWithCharactersBeforeDelimiter(t, newHarness)
+	})
 	t.Run("TestRead", func(t *testing.T) {
 		testRead(t, newHarness)
 	})
@@ -962,6 +974,86 @@ func testListDelimiters(t *testing.T, newHarness HarnessMaker) {
 	}
 }
 
+// testDirsWithCharactersBeforeDelimiter tests a case where there's
+// a directory on a pagination boundary that ends with a character that's
+// less than the delimiter.
+// See https://github.com/google/go-cloud/issues/3089.
+func testDirsWithCharactersBeforeDelimiter(t *testing.T, newHarness HarnessMaker) {
+	const keyPrefix = "blob-for-dirs-with-chars-before-delimiter/"
+	content := []byte("hello")
+
+	// The set of files to use for these tests.
+	keys := []string{
+		"testFile1",
+		"t/t/t",
+		"t-/t.",
+		"dir1/testFile1dir1",
+		"dir2/testFile1dir2",
+		"d",
+	}
+
+	// Note that "t-/" is before "t/". The delimiter is included in the
+	// alphabetical ordering.
+	want := []string{"d", "dir1/", "dir2/", "t-/", "t/", "testFile1"}
+
+	// Create blobs.
+	// We only create the blobs once, for efficiency and because there's
+	// no guarantee that after we create them they will be immediately returned
+	// from List. The very first time the test is run against a Bucket, it may be
+	// flaky due to this race.
+	ctx := context.Background()
+	h, err := newHarness(ctx, t)
+	if err != nil {
+		t.Fatal(err)
+	}
+	drv, err := h.MakeDriver(ctx)
+	if err != nil {
+		t.Fatal(err)
+	}
+	b := blob.NewBucket(drv)
+
+	// See if the blobs are already there.
+	iter := b.List(&blob.ListOptions{Prefix: keyPrefix})
+	found := iterToSetOfKeys(ctx, t, iter)
+	for _, key := range keys {
+		key = keyPrefix + key
+		if !found[key] {
+			if err := b.WriteAll(ctx, key, content, nil); err != nil {
+				b.Close()
+				t.Fatal(err)
+			}
+		}
+	}
+	defer b.Close()
+	defer h.Close()
+
+	opts := &blob.ListOptions{
+		Prefix:    keyPrefix,
+		Delimiter: "/",
+	}
+	// All page sizes should return the same end result.
+	for pageSize := 10; pageSize != 0; pageSize-- {
+		var got []string
+		objs, token, err := b.ListPage(ctx, blob.FirstPageToken, pageSize, opts)
+		for {
+			if err != nil {
+				t.Fatal(err)
+			}
+			for _, o := range objs {
+				key := strings.TrimPrefix(o.Key, keyPrefix)
+				got = append(got, key)
+			}
+			if token == nil {
+				break
+			}
+			objs, token, err = b.ListPage(ctx, token, pageSize, opts)
+		}
+		if !reflect.DeepEqual(want, got) {
+			t.Fatalf("For page size %d, got \n%v\nwant\n%v", pageSize, got, want)
+		}
+	}
+}
+
 func iterToSetOfKeys(ctx context.Context, t *testing.T, iter *blob.ListIterator) map[string]bool {
 	retval := map[string]bool{}
 	for {
@@ -1105,6 +1197,18 @@ func testRead(t *testing.T, newHarness HarnessMaker) {
 			if r.ModTime().IsZero() {
 				t.Errorf("got zero mod time, want non-zero")
 			}
+			// For tests that successfully read, recreate the io.Reader and
+			// test it with iotest.TestReader.
+			r, err = b.NewRangeReader(ctx, tc.key, tc.offset, tc.length, nil)
+			if err != nil {
+				t.Errorf("failed to recreate Reader: %v", err)
+				return
+			}
+			defer r.Close()
+			if err = iotest.TestReader(r, tc.want); err != nil {
+				t.Errorf("iotest.TestReader failed: %v", err)
+				return
+			}
 		})
 	}
 }
@@ -2210,7 +2314,9 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 
 	// Generate a signed URL for DELETE.
 	deleteURL, err := b.SignedURL(ctx, key, &blob.SignedURLOptions{Method: http.MethodDelete})
-	if err != nil {
+	if gcerrors.Code(err) == gcerrors.Unimplemented {
+		t.Log("DELETE URLs not supported, skipping")
+	} else if err != nil {
 		t.Fatal(err)
 	} else if deleteURL == "" {
 		t.Fatal("got empty DELETE url")
@@ -2231,20 +2337,24 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 	}
 	tests := []signedURLTest{
 		{"getURL", getURL, "", false},
-		{"deleteURL", deleteURL, "", false},
+	}
+	if deleteURL != "" {
+		tests = append(tests, signedURLTest{"deleteURL", deleteURL, "", false})
 	}
 	if putURLWithContentType != "" {
+		// Allowed content type should work.
+		// Different or empty content type should fail.
 		tests = append(tests, signedURLTest{"putURLWithContentType", putURLWithContentType, allowedContentType, true})
 		tests = append(tests, signedURLTest{"putURLWithContentType", putURLWithContentType, differentContentType, false})
 		tests = append(tests, signedURLTest{"putURLWithContentType", putURLWithContentType, "", false})
 	}
-	/*
-	 */
 	if putURLEnforcedAbsentContentType != "" {
+		// Empty content type should work, non-empty should fail.
 		tests = append(tests, signedURLTest{"putURLEnforcedAbsentContentType", putURLEnforcedAbsentContentType, "", true})
 		tests = append(tests, signedURLTest{"putURLEnforcedAbsentContentType", putURLEnforcedAbsentContentType, differentContentType, false})
 	}
 	if putURLWithoutContentType != "" {
+		// Empty content type should work.
 		tests = append(tests, signedURLTest{"putURLWithoutContentType", putURLWithoutContentType, "", true})
 		// From the SignedURLOptions docstring:
 		// If EnforceAbsentContentType is false and ContentType is the empty string,
@@ -2274,16 +2384,16 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 	}
 
 	// GET it. Try with all URLs, only getURL should work.
-	for _, test := range []struct {
-		urlDescription string
-		url            string
-		wantSuccess    bool
-	}{
-		{"deleteURL", deleteURL, false},
-		{"putURLWithoutContentType", putURLWithoutContentType, false},
-		{"getURLNoParams", getURLNoParams, false},
-		{"getURL", getURL, true},
-	} {
+	tests = nil
+	if deleteURL != "" {
+		tests = append(tests, signedURLTest{"deleteURL", deleteURL, "", false})
+	}
+	tests = append(tests, []signedURLTest{
+		{"putURLWithoutContentType", putURLWithoutContentType, "", false},
+		{"getURLNoParams", getURLNoParams, "", false},
+		{"getURL", getURL, "", true},
+	}...)
+	for _, test := range tests {
 		if resp, err := client.Get(test.url); err != nil {
 			t.Fatalf("GET with %s URL failed: %v", test.urlDescription, err)
 		} else {
@@ -2305,15 +2415,14 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 	}
 
 	// DELETE it. Try with all URLs, only deleteURL should work.
-	for _, test := range []struct {
-		urlDescription string
-		url            string
-		wantSuccess    bool
-	}{
-		{"getURL", getURL, false},
-		{"putURLWithoutContentType", putURLWithoutContentType, false},
-		{"deleteURL", deleteURL, true},
-	} {
+	tests = []signedURLTest{
+		{"getURL", getURL, "", false},
+		{"putURLWithoutContentType", putURLWithoutContentType, "", false},
+	}
+	if deleteURL != "" {
+		tests = append(tests, signedURLTest{"deleteURL", deleteURL, "", true})
+	}
+	for _, test := range tests {
 		req, err := http.NewRequest(http.MethodDelete, test.url, nil)
 		if err != nil {
 			t.Fatalf("failed to create DELETE HTTP request using %q: %v", test.urlDescription, err)
@@ -2324,22 +2433,24 @@ func testSignedURL(t *testing.T, newHarness HarnessMaker) {
 			defer resp.Body.Close()
 			success := resp.StatusCode >= 200 && resp.StatusCode < 300
 			if success != test.wantSuccess {
-				t.Fatalf("DELETE to %q got status code %d, want 2xx? %v", test.urlDescription, resp.StatusCode, test.wantSuccess)
 				gotBody, _ := ioutil.ReadAll(resp.Body)
 				t.Errorf(string(gotBody))
+				t.Fatalf("DELETE to %q got status code %d, want 2xx? %v", test.urlDescription, resp.StatusCode, test.wantSuccess)
 			}
 		}
 	}
 
 	// GET should fail now that the blob has been deleted.
-	if resp, err := client.Get(getURL); err != nil {
-		t.Errorf("GET after DELETE failed: %v", err)
-	} else {
-		defer resp.Body.Close()
-		if resp.StatusCode != 404 {
-			t.Errorf("GET after DELETE got status code %d, want 404", resp.StatusCode)
-			gotBody, _ := ioutil.ReadAll(resp.Body)
-			t.Errorf(string(gotBody))
+	if deleteURL != "" {
+		if resp, err := client.Get(getURL); err != nil {
+			t.Errorf("GET after DELETE failed: %v", err)
+		} else {
+			defer resp.Body.Close()
+			if resp.StatusCode != 404 {
+				t.Errorf("GET after DELETE got status code %d, want 404", resp.StatusCode)
+				gotBody, _ := ioutil.ReadAll(resp.Body)
+				t.Errorf(string(gotBody))
+			}
 		}
 	}
 }
