@@ -15,7 +15,7 @@
 // Package gcsblob provides a blob implementation that uses GCS. Use OpenBucket
 // to construct a *blob.Bucket.
 //
-// URLs
+// # URLs
 //
 // For blob.OpenBucket, gcsblob registers for the scheme "gs".
 // The default URL opener will set up a connection using default credentials
@@ -32,26 +32,27 @@
 // see URLOpener.
 // See https://gocloud.dev/concepts/urls/ for background information.
 //
-// Escaping
+// # Escaping
 //
 // Go CDK supports all UTF-8 strings; to make this work with services lacking
 // full UTF-8 support, strings must be escaped (during writes) and unescaped
 // (during reads). The following escapes are performed for gcsblob:
-//  - Blob keys: ASCII characters 10 and 13 are escaped to "__0x<hex>__".
-//    Additionally, the "/" in "../" is escaped in the same way.
+//   - Blob keys: ASCII characters 10 and 13 are escaped to "__0x<hex>__".
+//     Additionally, the "/" in "../" is escaped in the same way.
 //
-// As
+// # As
 //
 // gcsblob exposes the following types for As:
-//  - Bucket: *storage.Client
-//  - Error: *googleapi.Error
-//  - ListObject: storage.ObjectAttrs
-//  - ListOptions.BeforeList: *storage.Query
-//  - Reader: *storage.Reader
-//  - ReaderOptions.BeforeRead: **storage.ObjectHandle, *storage.Reader (if accessing both, must be in that order)
-//  - Attributes: storage.ObjectAttrs
-//  - CopyOptions.BeforeCopy: *CopyObjectHandles, *storage.Copier (if accessing both, must be in that order)
-//  - WriterOptions.BeforeWrite: **storage.ObjectHandle, *storage.Writer (if accessing both, must be in that order)
+//   - Bucket: *storage.Client
+//   - Error: *googleapi.Error
+//   - ListObject: storage.ObjectAttrs
+//   - ListOptions.BeforeList: *storage.Query
+//   - Reader: *storage.Reader
+//   - ReaderOptions.BeforeRead: **storage.ObjectHandle, *storage.Reader (if accessing both, must be in that order)
+//   - Attributes: storage.ObjectAttrs
+//   - CopyOptions.BeforeCopy: *CopyObjectHandles, *storage.Copier (if accessing both, must be in that order)
+//   - WriterOptions.BeforeWrite: **storage.ObjectHandle, *storage.Writer (if accessing both, must be in that order)
+//   - SignedURLOptions.BeforeSign: *storage.SignedURLOptions
 package gcsblob // import "gocloud.dev/blob/gcsblob"
 
 import (
@@ -82,6 +83,7 @@ import (
 	"gocloud.dev/gcerrors"
 	"gocloud.dev/gcp"
 	"gocloud.dev/internal/escape"
+	"gocloud.dev/internal/gcerr"
 	"gocloud.dev/internal/useragent"
 )
 
@@ -263,11 +265,13 @@ type Options struct {
 	// PrivateKey is the Google service account private key.
 	// Exactly one of PrivateKey or SignBytes must be non-nil to use SignedURL.
 	// See https://godoc.org/cloud.google.com/go/storage#SignedURLOptions.
+	// Deprecated: Use MakeSignBytes instead.
 	PrivateKey []byte
 
 	// SignBytes is a function for implementing custom signing.
 	// Exactly one of PrivateKey, SignBytes, or MakeSignBytes must be non-nil to use SignedURL.
 	// See https://godoc.org/cloud.google.com/go/storage#SignedURLOptions.
+	// Deprecated: Use MakeSignBytes instead.
 	SignBytes func([]byte) ([]byte, error)
 
 	// MakeSignBytes is a factory for functions that are being used in place of an empty SignBytes.
@@ -357,7 +361,7 @@ func (r *reader) As(i interface{}) bool {
 }
 
 func (b *bucket) ErrorCode(err error) gcerrors.ErrorCode {
-	if err == storage.ErrObjectNotExist {
+	if err == storage.ErrObjectNotExist || err == storage.ErrBucketNotExist {
 		return gcerrors.NotFound
 	}
 	if gerr, ok := err.(*googleapi.Error); ok {
@@ -480,6 +484,12 @@ func (b *bucket) Attributes(ctx context.Context, key string) (*driver.Attributes
 	if err != nil {
 		return nil, err
 	}
+	// GCS seems to unquote the ETag; restore them.
+	// It should be of the form "xxxx" or W/"xxxx".
+	eTag := attrs.Etag
+	if !strings.HasPrefix(eTag, "W/\"") && !strings.HasPrefix(eTag, "\"") && !strings.HasSuffix(eTag, "\"") {
+		eTag = fmt.Sprintf("%q", eTag)
+	}
 	return &driver.Attributes{
 		CacheControl:       attrs.CacheControl,
 		ContentDisposition: attrs.ContentDisposition,
@@ -487,9 +497,11 @@ func (b *bucket) Attributes(ctx context.Context, key string) (*driver.Attributes
 		ContentLanguage:    attrs.ContentLanguage,
 		ContentType:        attrs.ContentType,
 		Metadata:           attrs.Metadata,
+		CreateTime:         attrs.Created,
 		ModTime:            attrs.Updated,
 		Size:               attrs.Size,
 		MD5:                attrs.MD5,
+		ETag:               eTag,
 		AsFunc: func(i interface{}) bool {
 			p, ok := i.(*storage.ObjectAttrs)
 			if !ok {
@@ -697,7 +709,7 @@ func (b *bucket) SignedURL(ctx context.Context, key string, dopts *driver.Signed
 		numSigners++
 	}
 	if b.opts.GoogleAccessID == "" || numSigners != 1 {
-		return "", errors.New("to use SignedURL, you must call OpenBucket with a valid Options.GoogleAccessID and exactly one of Options.PrivateKey, Options.SignBytes, or Options.MakeSignBytes")
+		return "", gcerr.New(gcerr.Unimplemented, nil, 1, "gcsblob: to use SignedURL, you must call OpenBucket with a valid Options.GoogleAccessID and exactly one of Options.PrivateKey, Options.SignBytes, or Options.MakeSignBytes")
 	}
 
 	key = escapeKey(key)
@@ -712,6 +724,18 @@ func (b *bucket) SignedURL(ctx context.Context, key string, dopts *driver.Signed
 	if b.opts.MakeSignBytes != nil {
 		opts.SignBytes = b.opts.MakeSignBytes(ctx)
 	}
+	if dopts.BeforeSign != nil {
+		asFunc := func(i interface{}) bool {
+			v, ok := i.(**storage.SignedURLOptions)
+			if ok {
+				*v = opts
+			}
+			return ok
+		}
+		if err := dopts.BeforeSign(asFunc); err != nil {
+			return "", err
+		}
+	}
 	return storage.SignedURL(b.name, key, opts)
 }
 
