@@ -26,34 +26,38 @@ import (
 	"testing"
 	"time"
 
-	"github.com/golang/protobuf/proto"
-	"github.com/golang/protobuf/ptypes/timestamp"
 	"github.com/google/go-cmp/cmp"
 	"github.com/google/go-cmp/cmp/cmpopts"
+	"github.com/google/uuid"
 	"gocloud.dev/docstore"
 	ds "gocloud.dev/docstore"
 	"gocloud.dev/docstore/driver"
 	"gocloud.dev/gcerrors"
+	"google.golang.org/protobuf/proto"
+	tspb "google.golang.org/protobuf/types/known/timestamppb"
 )
 
+// ByteArray is an array of 2 bytes.
+type ByteArray [2]byte
+
 // CollectionKind describes the kind of testing collection to create.
 type CollectionKind int
 
 const (
-	// A collection with a single primary key field of type string named
+	// SingleKey is collection with a single primary key field of type string named
 	// drivertest.KeyField.
 	SingleKey CollectionKind = iota
 
-	// A collection that will consist entirely of HighScore structs (see below),
+	// TwoKey is a  collection that will consist entirely of HighScore structs (see below),
 	// whose two primary key fields are "Game" and "Player", both strings. Use
 	// drivertest.HighScoreKey as the key function.
 	TwoKey
 
-	// The collection should behave like a SingleKey collection, except
+	// AltRev is a collection that behaves like a SingleKey collection, except
 	// that the revision field should be drivertest.AlternateRevisionField.
 	AltRev
 
-	// The collection's documents will not have a revision field.
+	// NoRev is a collection whose documents will not have a revision field.
 	NoRev
 )
 
@@ -171,6 +175,7 @@ func RunConformanceTests(t *testing.T, newHarness HarnessMaker, ct CodecTester,
 	t.Run("Delete", func(t *testing.T) { withRevCollections(t, newHarness, testDelete) })
 	t.Run("Update", func(t *testing.T) { withRevCollections(t, newHarness, testUpdate) })
 	t.Run("Data", func(t *testing.T) { withCollection(t, newHarness, SingleKey, testData) })
+	t.Run("Proto", func(t *testing.T) { withCollection(t, newHarness, SingleKey, testProto) })
 	t.Run("MultipleActions", func(t *testing.T) { withRevCollections(t, newHarness, testMultipleActions) })
 	t.Run("GetQueryKeyField", func(t *testing.T) { withRevCollections(t, newHarness, testGetQueryKeyField) })
 	t.Run("SerializeRevision", func(t *testing.T) { withCollection(t, newHarness, SingleKey, testSerializeRevision) })
@@ -799,7 +804,7 @@ func testUpdate(t *testing.T, coll *ds.Collection, revField string) {
 			}
 			checkHasRevisionField(t, tc.doc, revField)
 			setRevision(tc.want, revision(got, revField), revField)
-			if diff := cmp.Diff(got, tc.want, cmpopts.IgnoreUnexported(timestamp.Timestamp{})); diff != "" {
+			if diff := cmp.Diff(got, tc.want, cmpopts.IgnoreUnexported(tspb.Timestamp{})); diff != "" {
 				t.Error(diff)
 			}
 		})
@@ -967,6 +972,9 @@ func testTypeDrivenDecode(t *testing.T, ct CodecTester) {
 		B:  true,
 		L:  []int{3, 4, 5},
 		A:  [2]int{6, 7},
+		A2: [2]int8{1, 2},
+		At: ByteArray{1, 2},
+		Uu: uuid.NameSpaceDNS,
 		M:  map[string]bool{"a": true, "b": false},
 		By: []byte{6, 7, 8},
 		P:  &s,
@@ -984,6 +992,9 @@ func testTypeDrivenDecode(t *testing.T, ct CodecTester) {
 		St: "foo",
 		B:  true,
 		L:  []int{3, 4, 5},
+		A:  [2]int{6, 7},
+		A2: [2]int8{6, 7},
+		At: ByteArray{1, 2},
 		M:  map[string]bool{"a": true, "b": false},
 		By: []byte{6, 7, 8},
 		P:  &s,
@@ -1146,6 +1157,9 @@ type docstoreRoundTrip struct {
 	By []byte
 	L  []int
 	A  [2]int
+	A2 [2]int8
+	At ByteArray
+	Uu uuid.UUID
 	M  map[string]bool
 	P  *string
 	T  time.Time
@@ -1163,6 +1177,9 @@ type nativeMinimal struct {
 	B  bool
 	By []byte
 	L  []int
+	A  [2]int
+	A2 [2]int8
+	At ByteArray
 	M  map[string]bool
 	P  *string
 	T  time.Time
@@ -1170,6 +1187,36 @@ type nativeMinimal struct {
 	LS []string
 }
 
+// testProto tests encoding/decoding of a document with protocol buffer
+// and pointer-to-protocol-buffer fields.
+func testProto(t *testing.T, _ Harness, coll *ds.Collection) {
+	ctx := context.Background()
+	type protoStruct struct {
+		Name             string `docstore:"name"`
+		Proto            tspb.Timestamp
+		PtrToProto       *tspb.Timestamp
+		DocstoreRevision interface{}
+	}
+	doc := &protoStruct{
+		Name:       "testing",
+		Proto:      tspb.Timestamp{Seconds: 42},
+		PtrToProto: &tspb.Timestamp{Seconds: 43},
+	}
+
+	err := coll.Create(ctx, doc)
+	if err != nil {
+		t.Fatal(err)
+	}
+	got := &protoStruct{}
+	err = coll.Query().Get(ctx).Next(ctx, got)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if diff := cmp.Diff(got, doc, cmpopts.IgnoreUnexported(tspb.Timestamp{})); diff != "" {
+		t.Error(diff)
+	}
+}
+
 // The following is the schema for the collection where the ID is composed from
 // multiple fields instead of one. It can be used for query testing.
 // It is loosely borrowed from the DynamoDB documentation.
@@ -1575,7 +1622,7 @@ func testMultipleActions(t *testing.T, coll *ds.Collection, revField string) {
 			got := gots[i]
 			want := clone(wants[i])
 			want[revField] = got[revField]
-			if !cmp.Equal(got, want, cmpopts.IgnoreUnexported(timestamp.Timestamp{})) {
+			if !cmp.Equal(got, want, cmpopts.IgnoreUnexported(tspb.Timestamp{})) {
 				t.Errorf("index #%d:\ngot  %v\nwant %v", i, got, want)
 			}
 		}
