@@ -16,8 +16,7 @@
 // and MongoDB-compatible services hosted on-premise or by cloud providers,
 // including Amazon DocumentDB and Azure Cosmos DB.
 //
-//
-// URLs
+// # URLs
 //
 // For docstore.OpenCollection, mongodocstore registers for the scheme "mongo".
 // The default URL opener will dial a Mongo server using the environment
@@ -26,26 +25,24 @@
 // see URLOpener.
 // See https://gocloud.dev/concepts/urls/ for background information.
 //
-//
-// Action Lists
+// # Action Lists
 //
 // mongodocstore uses the unordered BulkWrite call of the underlying driver for writes, and uses Find with a list of document IDs for Get.
 // (These implementation choices are subject to change.)
 // It calls the BeforeDo function once before each call to the underlying driver. The as function passed
 // to the BeforeDo function exposes the following types:
-//  - Gets: *options.FindOptions
-//  - writes: []mongo.WriteModel and *options.BulkWriteOptions
+//   - Gets: *options.FindOptions
+//   - writes: []mongo.WriteModel and *options.BulkWriteOptions
 //
-// As
+// # As
 //
 // mongodocstore exposes the following types for As:
-//  - Collection: *mongo.Collection
-//  - Query.BeforeQuery: *options.FindOptions or bson.D (the filter for Delete and Update queries)
-//  - DocumentIterator: *mongo.Cursor
-//  - Error: mongo.CommandError, mongo.BulkWriteError, mongo.BulkWriteException
+//   - Collection: *mongo.Collection
+//   - Query.BeforeQuery: *options.FindOptions or bson.D (the filter for Delete and Update queries)
+//   - DocumentIterator: *mongo.Cursor
+//   - Error: mongo.CommandError, mongo.BulkWriteError, mongo.BulkWriteException
 //
-//
-// Special Considerations
+// # Special Considerations
 //
 // MongoDB represents times to millisecond precision, while Go's time.Time type has
 // nanosecond precision. To save time.Times to MongoDB without loss of precision,
@@ -109,6 +106,7 @@ type collection struct {
 	opts          *Options
 }
 
+// Options holds various options.
 type Options struct {
 	// Lowercase all field names for document encoding, field selection, update
 	// modifications and queries.
@@ -294,7 +292,10 @@ func (c *collection) bulkFind(ctx context.Context, gets []*driver.Action, errs [
 func (c *collection) projectionDoc(fps [][]string) bson.D {
 	proj := bson.D{{Key: c.revisionField, Value: 1}}
 	for _, fp := range fps {
-		proj = append(proj, bson.E{Key: c.toMongoFieldPath(fp), Value: 1})
+		path := c.toMongoFieldPath(fp)
+		if path != c.revisionField {
+			proj = append(proj, bson.E{Key: path, Value: 1})
+		}
 	}
 	return proj
 }
