@@ -20,5 +20,5 @@ set -euo pipefail
 
 sed -e '/^#/d' -e '/^$/d' allmodules | awk '{print $1}' | while read -r path || [[ -n "$path" ]]; do
   echo "updating $path"
-  ( cd "$path" && go get -u &> /dev/null && go mod tidy &> /dev/null || echo "  FAILED! (some modules without code, like samples, are expected to fail)")
+  ( cd "$path" && go get -u ./... &> /dev/null && go mod tidy &> /dev/null || echo "  FAILED! (some modules without code, like samples, are expected to fail)")
 done
