@@ -16,7 +16,7 @@
 // PubSub. Use OpenTopic to construct a *pubsub.Topic, and/or OpenSubscription
 // to construct a *pubsub.Subscription.
 //
-// URLs
+// # URLs
 //
 // For pubsub.OpenTopic and pubsub.OpenSubscription, gcppubsub registers
 // for the scheme "gcppubsub".
@@ -31,22 +31,22 @@
 // So, when environment variable 'PUBSUB_EMULATOR_HOST' is set
 // driver connects to the specified emulator host by default.
 //
-// Message Delivery Semantics
+// # Message Delivery Semantics
 //
 // GCP Pub/Sub supports at-least-once semantics; applications must
 // call Message.Ack after processing a message, or it will be redelivered.
 // See https://godoc.org/gocloud.dev/pubsub#hdr-At_most_once_and_At_least_once_Delivery
 // for more background.
 //
-// As
+// # As
 //
 // gcppubsub exposes the following types for As:
-//  - Topic: *raw.PublisherClient
-//  - Subscription: *raw.SubscriberClient
-//  - Message.BeforeSend: *pb.PubsubMessage
-//  - Message.AfterSend: *string for the pb.PublishResponse.MessageIds entry corresponding to the message.
-//  - Message: *pb.PubsubMessage
-//  - Error: *google.golang.org/grpc/status.Status
+//   - Topic: *raw.PublisherClient
+//   - Subscription: *raw.SubscriberClient
+//   - Message.BeforeSend: *pb.PubsubMessage
+//   - Message.AfterSend: *string for the pb.PublishResponse.MessageIds entry corresponding to the message.
+//   - Message: *pb.PubsubMessage, *pb.ReceivedMessage
+//   - Error: *google.golang.org/grpc/status.Status
 package gcppubsub // import "gocloud.dev/pubsub/gcppubsub"
 
 import (
@@ -497,27 +497,33 @@ func (s *subscription) ReceiveBatch(ctx context.Context, maxMessages int) ([]*dr
 
 	ms := make([]*driver.Message, 0, len(resp.ReceivedMessages))
 	for _, rm := range resp.ReceivedMessages {
+		rm := rm
 		rmm := rm.Message
 		m := &driver.Message{
 			LoggableID: rmm.MessageId,
 			Body:       rmm.Data,
 			Metadata:   rmm.Attributes,
 			AckID:      rm.AckId,
-			AsFunc:     messageAsFunc(rmm),
+			AsFunc:     messageAsFunc(rmm, rm),
 		}
 		ms = append(ms, m)
 	}
 	return ms, nil
 }
 
-func messageAsFunc(pm *pb.PubsubMessage) func(interface{}) bool {
+func messageAsFunc(pm *pb.PubsubMessage, rm *pb.ReceivedMessage) func(interface{}) bool {
 	return func(i interface{}) bool {
-		p, ok := i.(**pb.PubsubMessage)
-		if !ok {
-			return false
+		ip, ok := i.(**pb.PubsubMessage)
+		if ok {
+			*ip = pm
+			return true
 		}
-		*p = pm
-		return true
+		rp, ok := i.(**pb.ReceivedMessage)
+		if ok {
+			*rp = rm
+			return true
+		}
+		return false
 	}
 }
 
@@ -547,8 +553,12 @@ func (s *subscription) SendNacks(ctx context.Context, ids []driver.AckID) error
 }
 
 // IsRetryable implements driver.Subscription.IsRetryable.
-func (s *subscription) IsRetryable(error) bool {
-	// The client handles retries.
+func (s *subscription) IsRetryable(err error) bool {
+	// The client mostly handles retries, but does not
+	// include DeadlineExceeded for some reason.
+	if s.ErrorCode(err) == gcerrors.DeadlineExceeded {
+		return true
+	}
 	return false
 }
 
